from matplotlib import rcParams
# This mpl style is from the UCSC BME163 class.
rcParams.update({
    'font.size'           : 8.0      ,
    'font.sans-serif'     : 'Arial'    ,
    'xtick.major.size'    : 2        ,
    'xtick.major.width'   : 0.75     ,
    'xtick.labelsize'     : 8.0      ,
    'xtick.direction'     : 'out'      ,
    'ytick.major.size'    : 2        ,
    'ytick.major.width'   : 0.75     ,
    'ytick.labelsize'     : 8.0      ,
    'ytick.direction'     : 'out'      ,
    'xtick.major.pad'     : 2        ,
    'xtick.minor.pad'     : 2        ,
    'ytick.major.pad'     : 2        ,
    'ytick.minor.pad'     : 2        ,
    'savefig.dpi'         : 601      ,
    'axes.linewidth'      : 0.75     ,
    'text.usetex'         : False     ,
    'text.latex.unicode'  : False     })
